import asyncio
import sys

from loguru import logger

from wizwalker import WizWalker, utils
from wizwalker.cli import WizWalkerConsole

logger.enable("wizwalker")
logger.remove(0)
logger.add("wizwalker_debug.log", level="DEBUG", rotation="10 MB")


async def run_console():
    walker = WizWalker()
    console = WizWalkerConsole(walker)
    await console.interact()


def main():
    if sys.platform != "win32":
        raise RuntimeError(f"This program is windows only, not {sys.platform}")

    asyncio.run(run_console())

    # app = WizWalker()
    #
    # # close WizWalker when app is closed
    # atexit.register(app.close)
    #
    # app.run()


def wiz_command():
    """
    called when someone uses the "wiz" console command
    """
    try:
        login_flag = sys.argv[1]
    except IndexError:
        utils.quick_launch()
        return

    if login_flag != "--login":
        print(f"{login_flag} is not a valid flag")
        return

    try:
        username = sys.argv[2]
    except IndexError:
        print("Missing username")
        return

    try:
        password = sys.argv[3]
    except IndexError:
        print("Missing password")
        return

    utils.quick_launch()

    # speed is the name :sunglasses:
    import time

    # this probably isn't enough time for some computers
    time.sleep(7)

    handles = utils.get_all_wizard_handles()
    newest_handle = sorted(handles)[-1]

    utils.wiz_login(newest_handle, username, password)


if __name__ == "__main__":
    main()



import json
import struct
from io import BytesIO

from loguru import logger

from wizwalker import utils

TYPE_TABLE = {
    "BYT": "b",
    "UBYT": "B",
    "SHRT": "<h",
    "USHRT": "<H",
    "INT": "<i",
    "UINT": "<I",
    "STR": "",
    "WSTR": "",
    "FLT": "<f",
    "DBL": "<d",
    "GID": "<Q",
}


class PacketProcesser:
    def __init__(self):
        self.message_structure = self.load_message_structure()

    @staticmethod
    def load_message_structure():
        json_data = utils.get_cache_folder() / "wizard_messages.json"

        try:
            return json.load(json_data.open())
        except FileNotFoundError:
            raise Exception("Messages not yet cached.")

    # Todo: comment what this does
    def process_packet_data(self, data):
        """
        Based on https://github.com/Joshsora/libki/wiki/Message-Structure
        """

        logger.debug(f"Parsing {data}")

        data_len = len(data)
        data = BytesIO(data)

        data.read(2)  # F00D header

        try:
            msg_length = struct.unpack(TYPE_TABLE["USHRT"], data.read(2))[0]
        except struct.error:
            print(f"Message len couldnt be read from: {data.read()}")
            return

        if msg_length != data_len - 2 - 2:
            logger.debug("Message inside message")
            pos = data.tell()
            real = data.read()
            next_msg_pos = real.find(b"\x0D\xF0")

            if next_msg_pos == 0:
                logger.debug(f"Message length issue {real=}")
                return

            self.process_packet_data(real[next_msg_pos:])
            data.seek(pos)

        is_control = struct.unpack(TYPE_TABLE["UBYT"], data.read(1))[0]

        if is_control:
            logger.debug("Ignoring control message")
            return

        op_code = struct.unpack(TYPE_TABLE["UBYT"], data.read(1))[0]
        data.read(2)  # undefined bytes
        service_id = struct.unpack(TYPE_TABLE["UBYT"], data.read(1))[0]

        # if service_id != 51:
        #     return

        message_id = struct.unpack(TYPE_TABLE["UBYT"], data.read(1))[0]

        logger.debug(
            f"message_len={msg_length} "
            f"service_id={service_id} "
            f"message_id={message_id} "
            f"is_control={is_control} "
            f"op_code={op_code}"
        )

        service_name, message_name, description, params = self.lookup_message_data(
            service_id, message_id
        )

        logger.debug(f"Got packet data: {service_name}: {message_name} ({description})")

        parsed_params = self.parse_params(data, params)

        return message_name, description, parsed_params

    def lookup_message_data(self, service_id: int, message_id: int):
        service_data = self.message_structure[str(service_id)]
        service_name = service_data["type"]

        message_data = service_data["messages"][str(message_id)]
        message_name = message_data["name"]
        message_description = message_data["description"]
        message_params = message_data["params"]

        return service_name, message_name, message_description, message_params

    @staticmethod
    def parse_params(data: BytesIO, params: list):
        parsed = {}
        for param in params:
            p_type = param["type"]
            p_name = param["name"]

            if p_type == "STR":
                # logger.debug("String type skipped")
                str_len = struct.unpack(TYPE_TABLE["USHRT"], data.read(2))[0]
                str_value = data.read(str_len)
                logger.debug(f"Got string param {str_len=} {str_value=}")
                parsed[p_name] = str_value
                continue

            p_format = TYPE_TABLE[p_type]
            p_size = struct.calcsize(p_format)

            logger.debug(f"{p_type=} {p_name=} {p_format=} {p_size=}")

            p_data = data.read(p_size)
            logger.debug(f"{p_data=}")

            try:
                p_value = struct.unpack(p_format, p_data)[0]

                logger.debug(f"{p_value=}")

                parsed[p_name] = p_value
            except Exception as e:
                logger.debug(f"Failed reading {p_name}")
                parsed[p_name] = "error"

        return parsed



import asyncio
import ctypes

from .constants import keycodes

user32 = ctypes.WinDLL("user32")


class KeyboardHandler:
    def __init__(self, window_handle):
        self.window_handle = window_handle

    async def send_key(self, key, seconds):
        # user32.PostMessageW(self.window_handle, 0x102, ord(key), 0)
        code = keycodes[key]
        send_task = asyncio.create_task(self.key_send_task(code))
        await asyncio.sleep(seconds)
        send_task.cancel()
        # https://docs.microsoft.com/en-us/windows/win32/inputdev/wm-keyup
        user32.PostMessageW(self.window_handle, 0x101, code, 0)

    async def key_send_task(self, code):
        while True:
            # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-sendmessagew
            # https://docs.microsoft.com/en-us/windows/win32/inputdev/wm-keydown
            user32.PostMessageW(self.window_handle, 0x100, code, 0)
            await asyncio.sleep(0.1)



import asyncio
import ctypes
import functools
import io
import math
import struct
import winreg
import zlib
from collections import namedtuple
from concurrent.futures.thread import ThreadPoolExecutor
from ctypes import WinDLL
from os import system as cmd
from pathlib import Path
from typing import Callable, Union
from xml.etree import ElementTree

import appdirs
from pymem.ptypes import RemotePointer

user32 = WinDLL("user32")
XYZ = namedtuple("XYZ", "x, y, z")


def _xyz_sub(self, other):
    return math.dist((self.x, self.y), (other.x, other.y))


XYZ.__sub__ = _xyz_sub


# Modified from https://github.com/Gorialis/jishaku/blob/master/jishaku/functools.py#L19
# This license covers the below function
# MIT License
#
# Copyright (c) 2020 Devon (Gorialis) R
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
def executor_function(sync_function: Callable):
    @functools.wraps(sync_function)
    async def sync_wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        internal_function = functools.partial(sync_function, *args, **kwargs)

        with ThreadPoolExecutor() as pool:
            return await loop.run_in_executor(pool, internal_function)

    return sync_wrapper


def get_wiz_install() -> Path:
    r"""
    Computer\HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Uninstall\
    {A9E27FF5-6294-46A8-B8FD-77B1DECA3021}

    <InstallLocation> value
    """
    reg = winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER)

    try:
        key = winreg.OpenKey(
            reg,
            r"Software\Microsoft\Windows\CurrentVersion\Uninstall\{A9E27FF5-6294-46A8-B8FD-77B1DECA3021}",
            0,
            winreg.KEY_READ,
        )
    except OSError:
        raise Exception("Wizard101 install not found, do you have it installed?")

    install_location = Path(winreg.QueryValueEx(key, "InstallLocation")[0]).absolute()
    return install_location


def start_wiz(location: Union[Path, str]):
    """
    <location>Bin\WizardGraphicalClient.exe -L login.us.wizard101.com 12000
    """
    # Todo: make a non shit way to do this
    cmd(
        f'start /D "{location}\\Bin" WizardGraphicalClient.exe -L login.us.wizard101.com 12000'
    )


def quick_launch():
    location = get_wiz_install()
    start_wiz(location)


def calculate_perfect_yaw(current_xyz: XYZ, target_xyz: XYZ) -> float:
    """
    Calculates the perfect yaw to reach target_xyz from current_xyz in a stright line
    """
    target_line = math.dist(
        (current_xyz.x, current_xyz.y), (target_xyz.x, target_xyz.y)
    )
    origin_line = math.dist(
        (current_xyz.x, current_xyz.y), (current_xyz.x, current_xyz.y - 1)
    )
    target_to_origin_line = math.dist(
        (target_xyz.x, target_xyz.y), (current_xyz.x, current_xyz.y - 1)
    )
    # target_angle = math.cos(origin_line / target_line)
    target_angle = math.acos(
        (pow(target_line, 2) + pow(origin_line, 2) - pow(target_to_origin_line, 2))
        / (2 * origin_line * target_line)
    )

    if target_xyz.x > current_xyz.x:
        # outside
        target_angle_degres = math.degrees(target_angle)
        perfect_yaw = math.radians(360 - target_angle_degres)
    else:
        # inside
        perfect_yaw = target_angle

    return perfect_yaw


def get_cache_folder() -> Path:
    app_name = "WizWalker"
    app_author = "StarrFox"
    cache_dir = Path(appdirs.user_cache_dir(app_name, app_author))

    if not cache_dir.exists():
        cache_dir.mkdir(parents=True)

    return cache_dir


def wiz_login(window_handle: int, username: str, password: str):
    def send_chars(chars: str):
        for char in chars:
            user32.PostMessageW(window_handle, 0x102, ord(char), 0)

    send_chars(username)
    # tab
    user32.PostMessageW(window_handle, 0x102, 9, 0)
    send_chars(password)
    # enter
    user32.PostMessageW(window_handle, 0x102, 13, 0)


def resolve_pointer(handle, base, offsets):
    last = base
    for offset in offsets:
        last = RemotePointer(handle, last.value + offset)

    return last.v.value


def get_all_wizard_handles() -> list:
    target_class = "Wizard Graphical Client"

    handles = []

    # callback takes a window handle and an lparam and returns true/false on if we should stop
    # iterating
    # https://docs.microsoft.com/en-us/previous-versions/windows/desktop/legacy/ms633498(v=vs.85)
    def callback(handle, _):
        class_name = ctypes.create_unicode_buffer(len(target_class))
        user32.GetClassNameW(handle, class_name, len(target_class) + 1)
        if target_class == class_name.value:
            handles.append(handle)

        # iterate all windows, (True)
        return 1

    # https://docs.python.org/3/library/ctypes.html#callback-functions
    enumwindows_func_type = ctypes.WINFUNCTYPE(
        ctypes.c_bool,  # return type
        ctypes.c_int,  # arg1 type
        ctypes.POINTER(ctypes.c_int),  # arg2 type
    )

    # Transform callback into a form we can pass to the dll
    callback = enumwindows_func_type(callback)

    # EnumWindows takes a callback every iteration is passed to
    # and an lparam
    # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-enumwindows
    user32.EnumWindows(callback, 0)

    return handles


def pharse_template_id_file(file_data: bytes) -> dict:
    if not file_data.startswith(b"BINd"):
        raise RuntimeError("No BINd id string")

    data = zlib.decompress(file_data[0xD:])

    total_size = len(data)
    data = io.BytesIO(data)

    data.seek(0x24)

    out = {}
    while data.tell() < total_size:
        size = ord(data.read(1)) // 2

        string = data.read(size).decode()
        data.read(8)  # unknown bytes

        # Little endian int
        entry_id = struct.unpack("<i", data.read(4))[0]

        data.read(0x10)  # next entry

        out[entry_id] = string

    return out


def pharse_message_file(file_data: bytes):
    decoded = file_data.decode(errors="ignore")
    root = ElementTree.fromstring(decoded)

    service_data = root.find("_ProtocolInfo").find("RECORD")
    service_id = int(service_data.find("ServiceID").text)
    pharsed_service_data = {
        "type": service_data.find("ProtocolType").text,
        "description": service_data.find("ProtocolDescription").text,
    }

    messages = root[1:]

    def msg_sorter(m):
        # Function to sort messages by
        return m[0].find("_MsgName").text

    parsed_msgs = {}
    for index, msg in enumerate(sorted(messages, key=msg_sorter), 1):
        # msg[0] is the RECORD element
        msg_data = msg[0]

        msg_name = msg_data.find("_MsgName").text
        msg_description = msg_data.find("_MsgDescription").text

        params = []

        for child in msg_data:
            # Message meta info starts with _
            if not child.tag.startswith("_"):
                params.append({"name": child.tag, "type": child.get("TYPE")})

        parsed_msgs[index] = {
            "name": msg_name,
            "description": msg_description,
            "params": params,
        }

    pharsed_service_data["messages"] = parsed_msgs

    return {service_id: pharsed_service_data}


def pharse_node_data(file_data: bytes) -> dict:
    """
    Converts data into a dict of node nums to points
    """
    entry_start = b"\xFE\xDB\xAE\x04"

    node_data = {}
    # no nodes
    if len(file_data) == 20:
        return node_data

    # header
    file_data = file_data[20:]

    last_start = 0
    while file_data:
        start = file_data.find(entry_start, last_start)
        if start == -1:
            break

        entry = file_data[start : start + 48 + 2]

        cords_data = entry[16 : 16 + (4 * 3)]
        x = struct.unpack("<f", cords_data[0:4])[0]
        y = struct.unpack("<f", cords_data[4:8])[0]
        z = struct.unpack("<f", cords_data[8:12])[0]

        node_num = entry[48 : 48 + 2]
        unpacked_num = struct.unpack("<H", node_num)[0]

        node_data[unpacked_num] = (x, y, z)

    return node_data



import asyncio
import math
from argparse import ArgumentParser

import aiofiles
from aioconsole import AsynchronousCli

from wizwalker import Wad, utils


# noinspection PyProtectedMember
class WizWalkerConsole(AsynchronousCli):
    def __init__(self, walker, **kwargs):
        commands = self.get_commands()
        super().__init__(commands, **kwargs)

        self.walker = walker

    def get_default_banner(self):
        msg = "WizWalker cli,\n"
        msg += "send list for the command list"
        return msg

    async def exit_command(self, reader, writer):
        writer.write("Closing wizwalker, hooks should be rewritten")
        await self.walker.close()
        await super().exit_command(reader, writer)

    def get_commands(self):
        commands = {}

        attach_parser = ArgumentParser(
            description="Attach to currently running wiz instances"
        )
        commands["attach"] = (self.attach_command, attach_parser)

        login_parser = ArgumentParser("login to a client")
        login_parser.add_argument("client_index", type=int)
        login_parser.add_argument("username", type=str)
        login_parser.add_argument("password", type=str)
        commands["login"] = (self.login_command, login_parser)

        hook_parser = ArgumentParser(description="Hooks into all processes")
        hook_parser.add_argument("--target_hook", type=str)
        hook_parser.add_argument("--list", action="store_true", dest="list_hooks")
        commands["hook"] = (self.hook_command, hook_parser)

        player_parser = ArgumentParser(description="Output various player information")
        commands["player"] = (self.player_command, player_parser)

        backpack_parser = ArgumentParser(
            description="Output various backpack information"
        )
        commands["backpack"] = (self.backpack_command, backpack_parser)

        quest_parser = ArgumentParser(description="Output various quest information")
        commands["quest"] = (self.quest_command, quest_parser)

        packet_parser = ArgumentParser(description="Output packet socket and buffer")
        commands["packet"] = (self.packet_command, packet_parser)

        watch_parser = ArgumentParser(description="Watch packets for information")
        commands["watch"] = (self.watch_command, watch_parser)

        zone_parser = ArgumentParser(description="Show current zone for each client")
        commands["zone"] = (self.zone_command, zone_parser)

        wad_parser = ArgumentParser(description="Extract a wad file")
        wad_parser.add_argument("wad_name", type=str)
        wad_parser.add_argument("file_name", type=str)
        wad_parser.add_argument("--output_name", type=str)
        commands["wad"] = (self.wad_command, wad_parser)

        cache_parser = ArgumentParser(description="Cache Wad data")
        commands["cache"] = (self.cache_command, cache_parser)

        wiz_parser = ArgumentParser(description="Start a Wizard101 instance")
        commands["wiz"] = (self.wiz_command, wiz_parser)

        send_parser = ArgumentParser(description="Send a key to all clients")
        send_parser.add_argument("key", type=str)
        send_parser.add_argument("seconds", type=float)
        commands["send"] = (self.send_command, send_parser)

        getspeed_parser = ArgumentParser(
            description="get number of units covered in 1 second"
        )
        commands["getspeed"] = (self.getspeed_command, getspeed_parser)

        teleport_parser = ArgumentParser(description="Teleport to a certain x,y,z")
        teleport_parser.add_argument("x", type=float)
        teleport_parser.add_argument("y", type=float)
        teleport_parser.add_argument("--z", type=float)
        teleport_parser.add_argument("--yaw", type=float)
        commands["teleport"] = (self.teleport_command, teleport_parser)

        goto_parser = ArgumentParser(description="go to a specific x y")
        goto_parser.add_argument("x", type=float)
        goto_parser.add_argument("y", type=float)
        commands["goto"] = (self.goto_command, goto_parser)

        return commands

    async def attach_command(self, _, writer):
        self.walker.get_clients()
        writer.write(f"Attached to {len(self.walker.clients)} clients\n")

    async def login_command(self, _, writer, client_index, username, password):
        try:
            self.walker.clients[client_index].login(username, password)
        except KeyError:
            writer.write(
                f"No client with index {client_index}, you only have {len(self.walker.clients)} clients\n"
            )
        else:
            writer.write(f"client-{client_index}: Logged in\n")

    async def hook_command(
        self, _, writer, target_hook: str = None, list_hooks: bool = False
    ):
        if list_hooks and target_hook:
            writer.write("You cannot list and hook at the same time smh\n")
            return

        if not self.walker.clients:
            writer.write("There are no attached clients to hook to\n")
            return

        if list_hooks:
            all_hooks = self.walker.clients[0].get_hooks()
            writer.write("all hooks:\n" + "\n".join(all_hooks) + "\n")
            return

        for idx, client in enumerate(self.walker.clients):
            if target_hook:
                await client.activate_hooks(target_hook.replace(" ", "_"))
                writer.write(f"client-{idx}: hooked {target_hook}\n")
            else:
                await client.activate_hooks()
                writer.write(f"client-{idx}: hooked all\n")

    async def player_command(self, _, writer):
        for idx, client in enumerate(self.walker.clients):
            writer.write(
                f"client-{idx}:\n"
                f"\txyz={await client.xyz()}\n"
                f"\thealth={await client.health()}\n"
                f"\tmana={await client.mana()}\n"
                f"\tpotions={await client.potions()}\n"
                f"\tgold={await client.gold()}\n"
                f"\tenergy={await client.energy()}\n"
                f"\tyaw={await client.yaw()}\n"
                f"\tplayer_base={hex(await client._memory.read_player_base())}\n"
                f"\tplayer_stat_base={hex(await client._memory.read_player_stat_base())}\n"
            )

    async def backpack_command(self, _, writer):
        for idx, client in enumerate(self.walker.clients):
            writer.write(
                f"client-{idx}: used_space={await client.backpack_space_used()} "
                f"total_space={await client.backpack_space_total()} "
                f"backpack_struct_addr={hex(await client._memory.read_backpack_stat_base())}\n"
            )

    async def quest_command(self, _, writer):
        for idx, client in enumerate(self.walker.clients):
            writer.write(f"client-{idx}: quest_xyz={await client.quest_xyz()}\n")

    async def packet_command(self, _, writer):
        for idx, client in enumerate(self.walker.clients):
            writer.write(
                f"client-{idx}: socket={await client._memory.read_packet_socket_discriptor()} "
                f"packet_buffer={await client._memory.read_packet_buffer()}\n"
            )

    async def watch_command(self, _, writer):
        for idx, client in enumerate(self.walker.clients):
            client.watch_packets()
            writer.write(f"client-{idx}: watching packets\n")

    async def zone_command(self, _, writer):
        for idx, client in enumerate(self.walker.clients):
            writer.write(f"client-{idx}: current zone={client.current_zone}\n")

    @staticmethod
    async def wad_command(_, writer, wad_name, file_name, output_name=None):
        wad = Wad(wad_name)
        file_data = await wad.get_file(file_name)

        if output_name is None:
            output_name = file_name

        async with aiofiles.open(output_name, "wb+") as fp:
            await fp.write(file_data)

        writer.write(f"Extracted {wad_name=}, {file_name=}\n")

    async def cache_command(self, _, writer):
        await self.walker.cache_data()
        writer.write("Cached\n")

    @staticmethod
    async def wiz_command(_, writer):
        utils.quick_launch()
        writer.write("Launched wizard101 instance\n")

    async def send_command(self, _, writer, key, seconds):
        if len(key) == 1:
            key = key.upper()

        for client in self.walker.clients:
            asyncio.create_task(client._keyboard.send_key(key, seconds))

        writer.write("Started\n")

    async def getspeed_command(self, _, writer):
        client = self.walker.clients[0]
        start = await client.xyz()
        await client._keyboard.send_key("W", 1)
        end = await client.xyz()
        distance = math.dist((start.x, start.y), (end.x, end.y))
        writer.write(f"from {start} to {end} covered {distance}\n")

    async def teleport_command(self, _, writer, x, y, z=None, yaw=None):
        for client in self.walker.clients:
            await client.teleport(x=x, y=y, z=z, yaw=yaw)

        writer.write("Teleported\n")

    async def goto_command(self, _, writer, x, y):
        for client in self.walker.clients:
            asyncio.create_task(client.goto(x, y))

        writer.write("Tasks started\n")



import socket

FOOD_START = b"\x0d\xf0"

# Todo: delete
class SocketListener:
    def __init__(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_IP)
        # self.socket.bind(("192.168.43.101", 0))
        self.socket.bind(("255.255.255.255", 0))
        self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_HDRINCL, 1)
        self.socket.ioctl(socket.SIO_RCVALL, socket.RCVALL_ON)

        self.is_listening = False

    def listen(self):
        self.is_listening = True

        while self.is_listening:
            raw_data = self.socket.recvfrom(65565)[0]

            if FOOD_START in raw_data:
                start = raw_data.find(FOOD_START)
                yield raw_data[start:]



import struct
from typing import Any, Tuple, Optional

import pymem
import pymem.pattern
from pymem.ressources.structure import MODULEINFO


# Modified to not handle memory protections
# This licence covers the below function
# MIT License
# Copyright (c) 2018 pymem
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
def scan_pattern_page(handle, address, pattern, mask):
    mbi = pymem.memory.virtual_query(handle, address)
    next_region = mbi.BaseAddress + mbi.RegionSize
    page_bytes = pymem.memory.read_bytes(handle, address, mbi.RegionSize)

    found = None
    for offset in range(0, (mbi.RegionSize - len(pattern)), 1):
        partial = page_bytes[offset : offset + len(pattern)]
        for x in range(len(pattern)):
            if mask[x] == "?":
                continue
            if mask[x] == "x" and not partial[x] == pattern[x]:
                break
        else:
            found = address + offset
            del page_bytes
            return None, found
    return next_region, found


# monkey patch to ignore memory protections
pymem.pattern.scan_pattern_page = scan_pattern_page


class MemoryHook:
    def __init__(self, memory_handler):
        self.memory_handler = memory_handler
        self.jump_original_bytecode = None

        self.hook_address = None
        self.jump_address = None

        self.jump_bytecode = None
        self.hook_bytecode = None

    def get_jump_address(self, pattern: bytes, mask: str, *, module=None) -> int:
        """
        gets the address to write jump at
        """
        if module:
            # noinspection PyTypeChecker
            jump_address = pymem.pattern.pattern_scan_module(
                self.memory_handler.process.process_handle, module, pattern, mask
            )

        else:
            # Todo: find faster way than scanning entire memory
            raise NotImplemented()

        return jump_address

    def get_hook_address(self, size: int) -> int:
        hook_address = self.memory_handler.process.allocate(size)
        return hook_address

    def get_jump_bytecode(self) -> bytes:
        """
        Gets the bytecode to write to the jump address
        """
        raise NotImplemented()

    def get_hook_bytecode(self) -> bytes:
        """
        Gets the bytecord to write to the hook address
        """
        raise NotImplemented()

    def get_pattern_and_mask(self) -> Tuple[bytes, Optional[str], Optional[MODULEINFO]]:
        raise NotImplemented()

    def hook(self) -> Any:
        """
        Writes jump_bytecode to jump address and hook bytecode to hook address
        """
        pattern, mask, module = self.get_pattern_and_mask()

        if mask is None:
            mask = "x" * len(pattern)

        self.jump_address = self.get_jump_address(pattern, mask, module=module)
        self.hook_address = self.get_hook_address(200)

        self.jump_bytecode = self.get_jump_bytecode()
        self.hook_bytecode = self.get_hook_bytecode()

        self.jump_original_bytecode = self.memory_handler.process.read_bytes(
            self.jump_address, len(self.jump_bytecode)
        )

        self.memory_handler.process.write_bytes(
            self.hook_address, self.hook_bytecode, len(self.hook_bytecode),
        )
        self.memory_handler.process.write_bytes(
            self.jump_address, self.jump_bytecode, len(self.jump_bytecode),
        )

    def unhook(self):
        """
        Deallocates hook memory and rewrites jump addr to it's origional code,
        also called when a client is closed
        """
        self.memory_handler.process.write_bytes(
            self.jump_address,
            self.jump_original_bytecode,
            len(self.jump_original_bytecode),
        )
        self.memory_handler.process.free(self.hook_address)


class PlayerHook(MemoryHook):
    def __init__(self, memory_handler):
        super().__init__(memory_handler)
        self.player_struct = None

    def get_pattern_and_mask(self) -> Tuple[bytes, Optional[str], Optional[MODULEINFO]]:
        module = pymem.process.module_from_name(
            self.memory_handler.process.process_handle, "WizardGraphicalClient.exe"
        )
        return b"\x8B\x48\x2C\x8B\x50\x30\x8B\x40\x34\xEB\x2A", None, module

    def set_player_struct(self):
        self.player_struct = self.memory_handler.process.allocate(4)

    def free_player_struct(self):
        if self.player_struct:
            self.memory_handler.process.free(self.player_struct)

    def get_jump_bytecode(self) -> bytes:
        """
        INJECT                           - E9 14458E01           - jmp 02730000
        WizardGraphicalClient.exe+A4BAEC - 90                    - nop
        """
        # distance = end - start
        distance = self.hook_address - self.jump_address
        relitive_jump = distance - 5  # size of this line
        packed_relitive_jump = struct.pack("<i", relitive_jump)
        return b"\xE9" + packed_relitive_jump + b"\x90"

    def get_hook_bytecode(self) -> bytes:
        self.set_player_struct()
        packed_addr = struct.pack("<i", self.player_struct)

        bytecode = (
            b"\x8B\xC8"  # mov ecx,eax
            b"\x81\xC1\x2C\x03\x00\x00"  # add ecx,0000032C { 812 }
            b"\x8B\x11"  # mov edx,[ecx]
            # check if player
            b"\x83\xFA\x08"  # cmp edx,08 { 8 }
            b"\x0F\x85\x05\x00\x00\x00"  # jne 314B0036 (relative jump 5 down)
            b"\xA3" + packed_addr +
            # original code
            b"\x8B\x48\x2C"  # mov ecx,[eax+2C]
            b"\x8B\x50\x30"  # mov edx,[eax+30]
        )

        return_addr = self.jump_address + len(self.jump_bytecode)

        relitive_return_jump = return_addr - (self.hook_address + len(bytecode)) - 5
        packed_relitive_return_jump = struct.pack("<i", relitive_return_jump)

        bytecode += (
            b"\xE9" + packed_relitive_return_jump
        )  # jmp WizardGraphicalClient.exe+A4BAED

        return bytecode

    def unhook(self):
        super().unhook()
        self.free_player_struct()


class PlayerStatHook(MemoryHook):
    def __init__(self, memory_handler):
        super().__init__(memory_handler)
        self.stat_addr = None

    def get_pattern_and_mask(self) -> Tuple[bytes, Optional[str], Optional[MODULEINFO]]:
        module = pymem.process.module_from_name(
            self.memory_handler.process.process_handle, "WizardGraphicalClient.exe"
        )
        return b"\x89\x48\x40\x74\x69", None, module

    def set_stat_addr(self):
        self.stat_addr = self.memory_handler.process.allocate(4)

    def free_stat_addr(self):
        if self.stat_addr:
            self.memory_handler.process.free(self.stat_addr)

    def get_jump_bytecode(self) -> bytes:
        # distance = end - start
        distance = self.hook_address - self.jump_address
        relitive_jump = distance - 5  # size of this line
        packed_relitive_jump = struct.pack("<i", relitive_jump)
        return b"\xE9" + packed_relitive_jump

    def get_hook_bytecode(self) -> bytes:
        self.set_stat_addr()
        packed_addr = struct.pack("<i", self.stat_addr)

        # registers gay
        tmp = self.memory_handler.process.allocate(4)
        packed_tmp = struct.pack("<i", tmp)

        bytecode = (
            # mov [tmp],edi
            b"\x89\x3D" + packed_tmp + b"\x8B\xF8"  # mov edi,eax
            b"\x89\x3D"
            + packed_addr  # mov [stat_addr],edi
            + b"\x8B\x3D"
            + packed_tmp
            +  # mov edi,[tmp]
            # original code
            b"\x89\x48\x40"  # mov [eax+40],ecx
        )

        # WizardGraphicalClient.exe+10060C3 + LEN(BYTECODE_LINES) - 6 (Length of this line)
        # 4 (length of target line) 1 (no idea)
        je_relitive_jump = (self.jump_address + 0x6E) - (
            self.hook_address + len(bytecode) + 6
        )
        packed_je_relitive_jump = struct.pack("<i", je_relitive_jump)

        bytecode += (
            b"\x0F\x84" + packed_je_relitive_jump
        )  # je WizardGraphicalClient.exe+10060C3

        jmp_relitive_jump = (self.jump_address + 0x5) - (
            self.hook_address + len(bytecode) + 5
        )
        packed_jmp_relitive_jump = struct.pack("<i", jmp_relitive_jump)

        bytecode += (
            b"\xE9" + packed_jmp_relitive_jump
        )  # jmp WizardGraphicalClient.exe+100605A"

        return bytecode

    def unhook(self):
        super().unhook()
        self.free_stat_addr()


class QuestHook(MemoryHook):
    def __init__(self, memory_handler):
        super().__init__(memory_handler)
        self.cord_struct = None

    def get_pattern_and_mask(self) -> Tuple[bytes, Optional[str], Optional[MODULEINFO]]:
        module = pymem.process.module_from_name(
            self.memory_handler.process.process_handle, "WizardGraphicalClient.exe"
        )
        return b"\xD9\x86\x1C\x08\x00\x00", "xxxx??", module

    def set_cord_struct(self):
        self.cord_struct = self.memory_handler.process.allocate(4)

    def free_cord_struct(self):
        if self.cord_struct:
            self.memory_handler.process.free(self.cord_struct)

    def get_jump_bytecode(self) -> bytes:
        # distance = end - start
        distance = self.hook_address - self.jump_address
        relitive_jump = distance - 5  # size of this line
        packed_relitive_jump = struct.pack("<i", relitive_jump)
        return b"\xE9" + packed_relitive_jump + b"\x90"

    def get_hook_bytecode(self) -> bytes:
        self.set_cord_struct()
        packed_addr = struct.pack("<i", self.cord_struct)  # little-endian int

        bytecode = (
            # original code
            b"\xD9\x86\x1C\x08\x00\00"  # original instruction one
            b"\x8D\xBE\x1C\x08\x00\00"  # original instruction two
            b"\x89\x35" + packed_addr
        )

        return_addr = self.jump_address + len(self.jump_bytecode)

        relitive_return_jump = return_addr - (self.hook_address + len(bytecode)) - 5
        packed_relitive_return_jump = struct.pack("<i", relitive_return_jump)

        bytecode += b"\xE9" + packed_relitive_return_jump

        return bytecode

    def unhook(self):
        super().unhook()
        self.free_cord_struct()


class BackpackStatHook(MemoryHook):
    def __init__(self, memory_handler):
        super().__init__(memory_handler)
        self.backpack_struct_addr = None

    def set_backpack_struct(self):
        self.backpack_struct_addr = self.memory_handler.process.allocate(4)

    def free_backpack_struct(self):
        self.memory_handler.process.free(self.backpack_struct_addr)

    def get_pattern_and_mask(self) -> Tuple[bytes, Optional[str], Optional[MODULEINFO]]:
        module = pymem.process.module_from_name(
            self.memory_handler.process.process_handle, "WizardGraphicalClient.exe"
        )
        return b"\xC7\x86\x70\x03\x00\x00\x00\x00\x00\x00\x74", None, module

    def get_jump_bytecode(self) -> bytes:
        # distance = end - start
        distance = self.hook_address - self.jump_address
        relitive_jump = distance - 5  # size of this line
        packed_relitive_jump = struct.pack("<i", relitive_jump)
        return b"\xE9" + packed_relitive_jump + b"\x0F\x1F\x44"

    def get_hook_bytecode(self) -> bytes:
        self.set_backpack_struct()
        packed_addr = struct.pack("<i", self.backpack_struct_addr)  # little-endian int

        ecx_tmp = self.memory_handler.process.allocate(4)
        packed_ecx_tmp = struct.pack("<i", ecx_tmp)

        bytecode = (
            b"\x89\x0D"
            + packed_ecx_tmp  # mov [02A91004],ecx { (0) }
            + b"\x8B\xCE"  # mov ecx,esi
            b"\x81\xC1\x70\x03\x00\x00"  # add ecx,00000370 { 880 }
            b"\x89\x0D"
            + packed_addr  # mov [packed_addr],ecx { (0) }
            + b"\x8B\x0D"
            + packed_ecx_tmp  # mov ecx,[02A91004] { (0) }
            +
            # original code
            b"\xC7\x86\x70\x03\x00\x00\x00\x00\x00\x00"  # mov [esi+00000370],00000000 { 0 }
        )

        # len of code at jump_address
        return_addr = self.jump_address + 10

        relitive_return_jump = return_addr - (self.hook_address + len(bytecode)) - 5
        packed_relitive_return_jump = struct.pack("<i", relitive_return_jump)

        bytecode += (
            b"\xE9" + packed_relitive_return_jump
        )  # jmp WizardGraphicalClient.exe+43ECD5

        return bytecode

    def unhook(self):
        super().unhook()
        self.free_backpack_struct()


class PacketHook(MemoryHook):
    def __init__(self, memory_handler):
        super().__init__(memory_handler)
        self.packet_buffer_addr = None
        self.socket_discriptor = None
        self.packet_buffer_len = None

        self.old_protection = None

    def set_packet_buffer_addr(self):
        self.packet_buffer_addr = self.memory_handler.process.allocate(4)

    def set_socket_discriptor(self):
        self.socket_discriptor = self.memory_handler.process.allocate(4)

    def set_packet_buffer_len(self):
        self.packet_buffer_len = self.memory_handler.process.allocate(4)

    def free_packet_buffer_addr(self):
        self.memory_handler.process.free(self.packet_buffer_addr)

    def free_socket_discriptor(self):
        self.memory_handler.process.free(self.socket_discriptor)

    def free_packet_buffer_len(self):
        self.memory_handler.process.free(self.packet_buffer_len)

    def get_jump_address(self, pattern: bytes, mask: str, *, module=None) -> int:
        """
        gets the address to write jump at
        """
        # exact location of wsock.recv + 2 which should never change
        jump_address = module.lpBaseOfDll + 0x1E32
        # I could read jump_address and compare to pattern but it's really unneeded
        return jump_address

    def get_pattern_and_mask(self) -> Tuple[bytes, Optional[str], Optional[MODULEINFO]]:
        module = pymem.process.module_from_name(
            self.memory_handler.process.process_handle, "WSOCK32.dll"
        )
        return b"\x8B\xFF\x55\x8B\xEC\x83\xEC\x10\x8B\x45\x10\x89\x45\xF0", None, module

    def get_jump_bytecode(self) -> bytes:
        # distance = end - start
        distance = self.hook_address - self.jump_address
        relitive_jump = distance - 5  # size of this line
        packed_relitive_jump = struct.pack("<i", relitive_jump)
        return b"\xE9" + packed_relitive_jump + b"\x90"

    def get_hook_bytecode(self) -> bytes:
        self.set_packet_buffer_addr()
        packed_buffer_addr = struct.pack("<i", self.packet_buffer_addr)

        self.set_socket_discriptor()
        packed_socket_discriptor = struct.pack("<i", self.socket_discriptor)

        self.set_packet_buffer_len()
        packed_buffer_len = struct.pack("<i", self.packet_buffer_len)

        ecx_tmp = self.memory_handler.process.allocate(4)
        packed_ecx_tmp = struct.pack("<i", ecx_tmp)

        # stack is stored here so we can restore it
        stack_buffer_tmp = self.memory_handler.process.allocate(16)
        packed_stack_buffer_tmp = struct.pack("<i", stack_buffer_tmp)
        packed_stack_buffer_tmp_4 = struct.pack("<i", stack_buffer_tmp + 0x4)
        packed_stack_buffer_tmp_8 = struct.pack("<i", stack_buffer_tmp + 0x8)
        packed_stack_buffer_tmp_12 = struct.pack("<i", stack_buffer_tmp + 0x12)

        # this is the function I'm hooking
        # int recv(
        #  SOCKET s,
        #  char * buf,
        #  int len,
        #  int flags);

        bytecode = (
            b"\x89\x0D" + packed_ecx_tmp +  # mov [ecx_tmp],ecx
            # reading args backwards (__stdcall)
            # int flags
            b"\x59"  # pop ecx
            b"\x89\x0D" + packed_stack_buffer_tmp +  # mov [stack_buffer],ecx
            # int len
            b"\x59"  # pop ecx
            b"\x89\x0D"
            + packed_stack_buffer_tmp_4  # mov [stack_buffer+4],ecx
            + b"\x89\x0D"
            + packed_buffer_len  # mov [buffer_len],ecx
            +
            # char * buf
            b"\x59"  # pop ecx
            b"\x89\x0D" + packed_stack_buffer_tmp_8 +  # mov [stack_buffer+8],ecx
            # move the buf pointer to our address
            b"\x89\x0D" + packed_buffer_addr +  # mov [buffer_addr],ecx
            # SOCKET s
            b"\x59"  # pop ecx
            b"\x89\x0D" + packed_stack_buffer_tmp_12 +  # mov [stack_buffer+12],ecx
            # move the socket descriptor to our addr
            b"\x89\x0D"
            + packed_socket_discriptor  # mov [socket_descriptor],ecx { (32768) }
            +
            # fix the stack for the rest of the recv function
            b"\x8B\x0D"
            + packed_stack_buffer_tmp_12  # mov ecx,[stack_buffer+12]
            + b"\x51"  # push ecx
            b"\x8B\x0D"
            + packed_stack_buffer_tmp_8  # mov ecx,[stack_buffer+8]
            + b"\x51"  # push ecx
            b"\x8B\x0D" + packed_stack_buffer_tmp_4 + b"\x51"  # see above
            b"\x8B\x0D" + packed_stack_buffer_tmp + b"\x51"
            b"\x8B\x0D" + packed_ecx_tmp +  # move ecx,[ecx_tmp]
            # original code
            b"\x55"  # push ebp
            b"\x8B\xEC"  # mov ebp,esp
            b"\x83\xEC\x10"  # sub esp,10
        )

        # WSOCK.recv+8 (jump_address is +2)
        return_addr = self.jump_address + 6

        relitive_return_jump = return_addr - (self.hook_address + len(bytecode)) - 5
        packed_relitive_return_jump = struct.pack("<i", relitive_return_jump)

        bytecode += b"\xE9" + packed_relitive_return_jump

        return bytecode

    def unhook(self):
        super().unhook()
        self.free_packet_buffer_addr()
        self.free_packet_buffer_len()
        self.free_socket_discriptor()



import asyncio
import re

ZONE_REGEX = re.compile(rb"\r\xf0..\x00\x00\x00\x00\x05l....([a-zA-Z/_]+)")


class PacketHookWatcher:
    def __init__(self, client, *, delay: float = 0.3):
        self.client = client
        # noinspection PyProtectedMember
        self.memory_handler = client._memory
        self.delay = delay
        self.watching = False

        # self.packet_processer = PacketProcesser()

    def start(self):
        self.watching = True
        asyncio.create_task(self.read_task())

    def stop(self):
        self.watching = False

    async def read_task(self):
        """
        Continusly reads the packet buffer every <self.delay> seconds
        """
        # don't drop packets on cancel
        while self.watching:
            buffer = await self.memory_handler.read_packet_buffer()
            await self.process_buffer(buffer)
            await asyncio.sleep(self.delay)

    async def process_buffer(self, buffer: bytes):
        """
        Splits packets and hands them to the processer
        """
        # Todo: make this work
        # packets = buffer.split(b"\x0D\xF0")
        # for packet in packets:
        #     packet = b"\x0D\xF0" + packet
        #     try:
        #         message_name, description, parsed_params = self.packet_processer.process_packet_data(packet)
        #     except:
        #         pass
        #     else:
        #         if "zone" in message_name.lower():
        #             print(f"Zone found, {parsed_params}")

        # I can't be bothered to parse the messages
        current_zone = ZONE_REGEX.search(buffer)
        if current_zone:
            self.client.current_zone = current_zone.group(1).decode("utf-8")



import asyncio
import functools
from collections import defaultdict

import pymem

from wizwalker import HookAlreadyActivated, utils, HookNotActive
from .hooks import PlayerHook, QuestHook, PlayerStatHook, BackpackStatHook, PacketHook


def uses_hook(hook: str):
    def decorator(function):
        @functools.wraps(function)
        async def wrapped(*args, **kwargs):
            memory_handler = args[0]
            if not memory_handler.active_hooks[hook]:
                raise HookNotActive(hook)

            return await function(*args, **kwargs)

        return wrapped

    return decorator


def register_hook(hook: str):
    def decorator(function):
        @functools.wraps(function)
        async def wrapped(*args, **kwargs):
            memory_handler = args[0]
            if memory_handler.active_hooks[hook]:
                raise HookAlreadyActivated(hook)

            memory_handler.set_hook_active(hook)

            return await function(*args, **kwargs)

        return wrapped

    return decorator


class MemoryHandler:
    """
    Handles the internal memory calls, not for public use
    """

    def __init__(self, pid: int):
        self.process = pymem.Pymem()
        self.process.open_process_from_id(pid)
        self.process.check_wow64()

        self.player_struct_addr = None
        self.quest_struct_addr = None
        self.player_stat_addr = None
        self.backpack_stat_addr = None
        self.packet_socket_discriptor_addr = None
        self.packet_buffer_addr = None
        self.packet_buffer_len = None

        self.hooks = []
        self.active_hooks = defaultdict(lambda: False)

    @utils.executor_function
    def close(self):
        """
        Closes MemoryHandler, closing all hooks
        """
        for hook in self.hooks:
            try:
                hook.unhook()
            except pymem.exception.MemoryWriteError:
                pass

    def set_hook_active(self, hook):
        self.active_hooks[hook] = True

    def hook_functs(self):
        hooks = {}
        # Couldn't get anything else working for this, other than manually setting
        # some attr on each hook method
        for thing in dir(self):
            if thing.startswith("hook_") and not any(
                thing.endswith(i) for i in ("all", "functs")
            ):
                hooks[thing.replace("hook_", "")] = getattr(self, thing)

        return hooks

    def is_hook_active(self, hook):
        return self.active_hooks[hook]

    @uses_hook("player_struct")
    @utils.executor_function
    def read_player_base(self):
        return self.process.read_int(self.player_struct_addr)

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_stat_base(self):
        return self.process.read_int(self.player_stat_addr)

    @uses_hook("backpack_stat_struct")
    @utils.executor_function
    def read_backpack_stat_base(self):
        return self.process.read_int(self.backpack_stat_addr)

    @uses_hook("quest_struct")
    @utils.executor_function
    def read_quest_base(self):
        return self.process.read_int(self.quest_struct_addr)

    @uses_hook("player_struct")
    @utils.executor_function
    def read_xyz(self):
        player_struct = self.process.read_int(self.player_struct_addr)
        try:
            x = self.process.read_float(player_struct + 0x2C)
            y = self.process.read_float(player_struct + 0x30)
            z = self.process.read_float(player_struct + 0x34)
        except pymem.exception.MemoryReadError:
            return None
        else:
            return utils.XYZ(x, y, z)

    @uses_hook("player_struct")
    @utils.executor_function
    def set_xyz(self, *, x=None, y=None, z=None):
        player_struct = self.process.read_int(self.player_struct_addr)
        try:
            if x is not None:
                self.process.write_float(player_struct + 0x2C, x)
            if y is not None:
                self.process.write_float(player_struct + 0x30, y)
            if z is not None:
                self.process.write_float(player_struct + 0x34, z)
        except pymem.exception.MemoryWriteError:
            return False
        else:
            return True

    @uses_hook("player_struct")
    @utils.executor_function
    def read_player_yaw(self):
        player_struct = self.process.read_int(self.player_struct_addr)
        try:
            return self.process.read_float(player_struct + 0x40)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_struct")
    @utils.executor_function
    def set_player_yaw(self, yaw):
        player_struct = self.process.read_int(self.player_struct_addr)
        try:
            self.process.write_float(player_struct + 0x40, yaw)
        except pymem.exception.MemoryWriteError:
            return False
        else:
            return True

    @uses_hook("player_struct")
    @utils.executor_function
    def read_player_pitch(self):
        player_struct = self.process.read_int(self.player_struct_addr)
        try:
            return self.process.read_float(player_struct + 0x38)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_struct")
    @utils.executor_function
    def set_player_pitch(self, pitch):
        player_struct = self.process.read_int(self.player_struct_addr)
        try:
            self.process.write_float(player_struct + 0x38, pitch)
        except pymem.exception.MemoryWriteError:
            return False
        else:
            return True

    @uses_hook("player_struct")
    @utils.executor_function
    def read_player_roll(self):
        player_struct = self.process.read_int(self.player_struct_addr)
        try:
            return self.process.read_float(player_struct + 0x3C)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_struct")
    @utils.executor_function
    def set_player_roll(self, roll):
        player_struct = self.process.read_int(self.player_struct_addr)
        try:
            self.process.write_float(player_struct + 0x3C, roll)
        except pymem.exception.MemoryWriteError:
            return False
        else:
            return True

    @uses_hook("quest_struct")
    @utils.executor_function
    def read_quest_xyz(self):
        quest_struct = self.process.read_int(self.quest_struct_addr)
        x = self.process.read_float(quest_struct + 0x81C)
        y = self.process.read_float(quest_struct + 0x81C + 0x4)
        z = self.process.read_float(quest_struct + 0x81C + 0x8)

        return utils.XYZ(x, y, z)

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_health(self):
        stat_addr = self.process.read_int(self.player_stat_addr)
        try:
            return self.process.read_int(stat_addr + 0x40)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_mana(self):
        stat_addr = self.process.read_int(self.player_stat_addr)
        try:
            return self.process.read_int(stat_addr + 0x10 + 0x40)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_energy(self):
        stat_addr = self.process.read_int(self.player_stat_addr)
        try:
            return self.process.read_int(stat_addr + 0x3C)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_potions(self):
        stat_addr = self.process.read_int(self.player_stat_addr)
        try:
            # this is a float for some reason
            return int(self.process.read_float(stat_addr + 0x2C + 0x40))
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_gold(self):
        stat_addr = self.process.read_int(self.player_stat_addr)
        try:
            return self.process.read_int(stat_addr + 0x4 + 0x40)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("backpack_stat_struct")
    @utils.executor_function
    def read_player_backpack_used(self):
        backpack_addr = self.process.read_int(self.backpack_stat_addr)
        try:
            return self.process.read_int(backpack_addr)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("backpack_stat_struct")
    @utils.executor_function
    def read_player_backpack_total(self):
        backpack_addr = self.process.read_int(self.backpack_stat_addr)
        try:
            return self.process.read_int(backpack_addr + 0x4)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("packet_recv")
    @utils.executor_function
    def read_packet_socket_discriptor(self):
        try:
            return self.process.read_bytes(self.packet_socket_discriptor_addr, 20)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("packet_recv")
    @utils.executor_function
    def read_packet_buffer(self):
        buffer_addr = self.process.read_int(self.packet_buffer_addr)
        buffer_len = self.process.read_int(self.packet_buffer_len)
        try:
            return self.process.read_bytes(buffer_addr, buffer_len)
        except pymem.exception.MemoryReadError:
            return None

    async def hook_all(self):
        hooks = []
        for hook in self.hook_functs().values():
            # Need the coroutine objects
            hooks.append(hook())

        # return_exceptions=True will make all exceptions return as results
        return await asyncio.gather(*hooks, return_exceptions=True)

    @register_hook("player_struct")
    @utils.executor_function
    def hook_player_struct(self):
        player_hook = PlayerHook(self)
        player_hook.hook()

        self.hooks.append(player_hook)
        self.player_struct_addr = player_hook.player_struct

    @register_hook("quest_struct")
    @utils.executor_function
    def hook_quest_struct(self):
        quest_hook = QuestHook(self)
        quest_hook.hook()

        self.hooks.append(quest_hook)
        self.quest_struct_addr = quest_hook.cord_struct

    @register_hook("player_stat_struct")
    @utils.executor_function
    def hook_player_stat_struct(self):
        player_stat_hook = PlayerStatHook(self)
        player_stat_hook.hook()

        self.hooks.append(player_stat_hook)
        self.player_stat_addr = player_stat_hook.stat_addr

    @register_hook("backpack_stat_struct")
    @utils.executor_function
    def hook_backpack_stat_struct(self):
        backpack_stat_hook = BackpackStatHook(self)
        backpack_stat_hook.hook()

        self.hooks.append(backpack_stat_hook)
        self.backpack_stat_addr = backpack_stat_hook.backpack_struct_addr

    @register_hook("packet_recv")
    @utils.executor_function
    def hook_packet_recv(self):
        packet_recv_hook = PacketHook(self)
        packet_recv_hook.hook()

        self.hooks.append(packet_recv_hook)
        self.packet_socket_discriptor_addr = packet_recv_hook.socket_discriptor
        self.packet_buffer_addr = packet_recv_hook.packet_buffer_addr
        self.packet_buffer_len = packet_recv_hook.packet_buffer_len



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wizwalker',
 'wizwalker.cli',
 'wizwalker.combat',
 'wizwalker.packets',
 'wizwalker.windows',
 'wizwalker.windows.memory']

package_data = \
{'': ['*']}

install_requires = \
['aioconsole>=0.2.1,<0.3.0',
 'aiofiles>=0.5.0,<0.6.0',
 'appdirs>=1.4.4,<2.0.0',
 'loguru>=0.5.1,<0.6.0',
 'pymem==1.3']

entry_points = \
{'console_scripts': ['wiz = wizwalker.__main__:wiz_command',
                     'wizwalker = wizwalker.__main__:main']}

setup_kwargs = {
    'name': 'wizwalker',
    'version': '0.12.2',
    'description': 'Automation bot for wizard101',
    'long_description': "# In development\n# WizWalker\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nWizard101 quest bot scripting api and application\n\n## documentation\nyou can find the documentation [here](https://starrfox.github.io/wizwalker/)\n\nyou can download these from the gh-pages branch if desired\n\n## install\ndownload latest release from [here](https://github.com/StarrFox/WizWalker/releases)\nor install from pypi `pip install -U wizwalker`\n\n## discord\njoin the offical discord [here](https://discord.gg/JHrdCNK)\n\n## development install\nThis package uses [poetry](https://python-poetry.org/)\n```shell script\n$ poetry install\n```\n\n## running\nShell may need admin perms\n```shell script\n$ poetry shell\n$ py -m wizwalker\n```\n\n## building\nYou'll need the dev install (see above) for this to work\n\n### exe\n```shell script\n# Admin if needed\n$ pyinstaller -F --uac-admin --name WizWalker wizwalker/__main__.py\n# Normal\n$ pyinstaller -F --name WizWalker wizwalker/__main__.py\n```\n\n### wheel and source\n```shell script\n$ poetry build\n```\n\n### Docs\n```shell script\n$ cd docs\n$ make html\n```\n\n## console commands\nwizwalker: Runs the wizwalker bot\n\nwiz: Starts a Wizard101 instance\n\n## project goals in order of importance\n0. ~~basic info by memory~~\n1. able to determine current quest\n2. ~~teleportion mode~~\n3. ~~info by memory~~\n4. ~~info by packet~~\n5. ~~able to walk~~\n6. able to combat\n7. ~~cli for end users~~\n8. gui for end users\n",
    'author': 'StarrFox',
    'author_email': 'starrfox6312@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/StarrFox/wizwalker',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import struct
import zlib
from collections import namedtuple

import aiofiles

from .utils import executor_function, get_wiz_install

wad_file_info = namedtuple("wad_file_info", "offset, size, is_zip, crc, unzipped_size")


class Wad:
    def __init__(self, name: str):
        if not name.endswith(".wad"):
            name += ".wad"

        self.name = name
        self.file_path = get_wiz_install() / "Data" / "GameData" / self.name
        if not self.file_path.exists():
            raise RuntimeError(f"{self.name} not found.")

        self.journal = {}
        self.refreshed_once = False

    def __repr__(self):
        return f"<Wad {self.name=}>"

    async def total_size(self):
        """Total size of the files in this wad"""
        total = 0
        for file_info in self.journal.values():
            total += file_info.size

        return total

    # This method runs in a thread bc it is very slow
    @executor_function
    def refresh_journal(self):
        self.refreshed_once = True

        with open(self.file_path, "rb") as fp:
            # KIWAD id string
            fp.seek(5)
            version = struct.unpack("<l", fp.read(4))[0]
            file_num = struct.unpack("<l", fp.read(4))[0]

            if version >= 2:
                fp.read(1)

            for _ in range(file_num):
                offset = struct.unpack("<l", fp.read(4))[0]
                size = struct.unpack("<l", fp.read(4))[0]
                zsize = struct.unpack("<l", fp.read(4))[0]
                is_zip = struct.unpack("?", fp.read(1))[0]
                crc = struct.unpack("<l", fp.read(4))[0]
                name_length = struct.unpack("<l", fp.read(4))[0]
                name = (fp.read(name_length)).decode("utf-8")[:-1]

                self.journal[name] = wad_file_info(
                    offset=offset,
                    size=size,
                    is_zip=is_zip,
                    crc=crc,
                    unzipped_size=zsize,
                )

    async def get_file(self, name: str) -> bytes:
        """Get the data contents of the named file"""
        if not self.refreshed_once:
            await self.refresh_journal()

        target_file = self.journal.get(name)

        if not target_file:
            raise RuntimeError(f"File {name} not found.")

        async with aiofiles.open(self.file_path, "rb") as fp:
            await fp.seek(target_file.offset)
            raw_data = await fp.read(target_file.size)

            if target_file.is_zip:
                data = zlib.decompress(raw_data)

            else:
                data = raw_data

        return data

    async def get_file_info(self, name: str) -> wad_file_info:
        """
        Gets a wad_file_info for a file
        this is the same as journal[name]
        """
        if not self.refreshed_once:
            await self.refresh_journal()

        try:
            file_info = self.journal[name]
        except KeyError:
            raise RuntimeError(f"File {name} not found")
        else:
            return file_info



import ctypes.wintypes
from functools import cached_property
from typing import Optional

from . import utils
from .packets import PacketHookWatcher
from .windows import KeyboardHandler, MemoryHandler, user32


WIZARD_SPEED = 580


class Client:
    """
    Represents a connected wizard client

    Args:
        window_handle: A handle to the window this client connects to
    """

    def __init__(self, window_handle: int):
        self.window_handle = window_handle
        self._keyboard = KeyboardHandler(window_handle)
        self._memory = MemoryHandler(self.process_id)
        self.current_zone = None

        self.packet_watcher = PacketHookWatcher(self)

    def __repr__(self):
        return f"<Client {self.window_handle=} {self.process_id=}>"

    async def close(self):
        """
        Closes this client; unhooking all active hooks
        """
        await self._memory.close()

    @cached_property
    def process_id(self) -> int:
        """
        This client's PID

        Returns:
            The pid
        """
        # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getwindowthreadprocessid
        pid = ctypes.wintypes.DWORD()
        user32.GetWindowThreadProcessId(self.window_handle, ctypes.byref(pid))
        return pid.value

    def login(self, username: str, password: str):
        """
        Login to a client that is at the login screen

        Args:
            username: The username to login with
            password: The password to login with
        """
        utils.wiz_login(self.window_handle, username, password)

    def watch_packets(self):
        """
        Start Watching packets for information
        """
        self.packet_watcher.start()

    def get_hooks(self) -> list:
        """
        return a list of hook_names from the underlying MemoryHandler

        Returns:
            A list of avalible hooks
        """
        return self._memory.hook_functs().keys()

    async def activate_hooks(self, *hook_names: Optional[str]):
        """
        Activate a number of hooks or pass None/no args to activate all

        Args:
            hook_names: The hooks to activate

        Examples:
            .. code-block:: py

                # activates player_struct and player_stat_struct
                activate_hooks("player_struct", "player_stat_struct")

                # activates all hooks
                activate_hooks()

        """
        if not hook_names:
            await self._memory.hook_all()

        else:
            for hook_name in hook_names:
                hook = getattr(self._memory, "hook_" + hook_name, None)
                if not hook:
                    raise ValueError(f"{hook_name} is not a valid hook")

                await hook()

    async def goto(self, x: float, y: float, *, use_nodes: bool = False):
        """
        Moves the player to a specific x and y

        Args:
            x: X to move to
            y: Y to move to
            use_nodes: If node date should be used, currently WIP
        """
        if use_nodes is False:
            await self._to_point(x, y)
        else:
            raise NotImplemented("WIP")

    async def _to_point(self, x, y):
        current_xyz = await self.xyz()
        target_xyz = utils.XYZ(x, y, current_xyz.z)
        distance = current_xyz - target_xyz
        move_seconds = distance / WIZARD_SPEED
        yaw = utils.calculate_perfect_yaw(current_xyz, target_xyz)

        await self.set_yaw(yaw)
        await self._keyboard.send_key("W", move_seconds)

    async def teleport(
        self, *, x: float = None, y: float = None, z: float = None, yaw: float = None
    ) -> bool:
        """
        Teleport the player to a set x, y, z

        Args:
            x: X to teleport to
            y: Y to teleport to
            z: Z to teleport to
            yaw: yaw to set

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            True if telelporting succseeded, False otherwise
        """
        res = await self._memory.set_xyz(x=x, y=y, z=z)

        if yaw is not None:
            await self._memory.set_player_yaw(yaw)

        return res

    async def xyz(self) -> Optional[utils.XYZ]:
        """
        Player xyz

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            XYZ namedtuple or None if hooked function hasn't run yet
        """
        return await self._memory.read_xyz()

    async def yaw(self) -> Optional[float]:
        """
        Player yaw

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            yaw float or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_yaw()

    async def set_yaw(self, yaw: float) -> bool:
        """
        Set the player yaw

        Raises:
            RuntimeError: player_struct hook not active

        Return:
            True if value was set, False otherwise
        """
        return await self._memory.set_player_yaw(yaw)

    async def roll(self) -> Optional[float]:
        """
        Player roll

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            roll float or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_roll()

    async def set_roll(self, roll: float) -> bool:
        """
        Set the player roll

        Raises:
            RuntimeError: player_struct hook not active

        Return:
            True if value was set, False otherwise
        """
        return await self._memory.set_player_roll(roll)

    async def pitch(self) -> Optional[float]:
        """
        Player pitch

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            pitch float or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_pitch()

    async def set_pitch(self, pitch: float) -> bool:
        """
        Set the player pitch

        Raises:
            RuntimeError: player_struct hook not active

        Return:
            True if value was set, False otherwise
        """
        return await self._memory.set_player_roll(pitch)

    async def quest_xyz(self) -> Optional[utils.XYZ]:
        """
        Quest xyz

        Raises:
            RuntimeError: quest_struct hook not active

        Return:
            XYZ namedtuple or None if hooked function hasn't run yet
        """
        return await self._memory.read_quest_xyz()

    async def health(self) -> Optional[int]:
        """
        Player health

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            health int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_health()

    async def mana(self) -> Optional[int]:
        """
        Player mana

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            mana int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_mana()

    async def energy(self) -> Optional[int]:
        """
        Player energy

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            energy int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_energy()

    async def potions(self) -> Optional[int]:
        """
        Player full potions

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            # of potions int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_potions()

    async def gold(self) -> Optional[int]:
        """
        Player gold

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            gold int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_gold()

    async def backpack_space_used(self) -> Optional[int]:
        """
        Player backpack used space

        Raises:
            RuntimeError: backpack_struct hook not active

        Return:
            used backpack space int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_backpack_used()

    async def backpack_space_total(self) -> Optional[int]:
        """
        Player backpack total space

        Raises:
            RuntimeError: backpack_struct hook not active

        Return:
            backpack total space int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_backpack_total()
